@@ -2,14 +2,14 @@ from __future__ import annotations
 
 from ._harmony_integrate import harmony_integrate
 from ._hvg import highly_variable_genes
-from ._neighbors import neighbors
+from ._neighbors import bbknn, neighbors
 from ._normalize import log1p, normalize_pearson_residuals, normalize_total
 from ._pca import pca
+from ._qc import calculate_qc_metrics
 from ._regress_out import regress_out
 from ._scale import scale
 from ._scrublet import scrublet, scrublet_simulate_doublets
 from ._simple import (
-    calculate_qc_metrics,
     filter_cells,
     filter_genes,
     filter_highly_variable,
