@@ -1,12 +1,10 @@
 from __future__ import annotations
 
 import math
+from collections.abc import Iterable, Mapping, Sequence
 from itertools import product
 from typing import (
-    Iterable,
     Literal,
-    Mapping,
-    Sequence,
 )
 
 import cupy as cp
@@ -240,7 +238,7 @@ def ligrec(
             the above mentioned dict.
 
     NaN p-values mark combinations for which the mean expression of one of the \
-    interacting components was 0 or it didnt pass the threshold percentage of \
+    interacting components was 0 or it didn't pass the threshold percentage of \
     cells being expressed within a given cluster.
     """
     # Get and Check interactions
@@ -345,7 +343,13 @@ def ligrec(
             return complexes[0]
 
         df = data[complexes].mean()
-        return str(df.index[df.argmin()])
+        try:
+            return str(df.index[df.argmin()])
+        except ValueError as e:
+            if "attempt to get argmin of an empty sequence" in str(e):
+                return str(df.index[0])
+            else:
+                raise ValueError(e)
 
     if complex_policy == "min":
         interactions[SOURCE] = interactions[SOURCE].apply(find_min_gene_in_complex)
